/**
  *  \file planetscentral/dbclient/reconnect.h
  *  \brief Database Client: Automatic Reconnection
  */
#ifndef C2SERVER_PLANETSCENTRAL_DBCLIENT_RECONNECT_H
#define C2SERVER_PLANETSCENTRAL_DBCLIENT_RECONNECT_H

#include "planetscentral/osutil.h"
#include "cpluslib/string.h"

class Connection;

/** Automatic Reconnection and Locking. Multiple users using a single
    Connection object must lock their access to that object to avoid
    buffering problems. In addition, a connection might be closed when
    the opposite server stops and restarts. This class, together with
    ReconnectingGuard, provides the necessary logic.
    - before using the Connection, call enter(), or construct a
      ReconnectingGuard.
    - when finished with the Connection, call leave(), or destroy the
      ReconnectingGuard. */
class ReconnectingConnection {
 public:
    ReconnectingConnection(Connection& conn, Mutex& outputMutex, string_t host, string_t port);

    void enter();
    void leave();

    void connectRetry();

    void setHost(string_t h);
    void setPort(string_t p);

    string_t getHost() const;
    string_t getPort() const;
    
 private:
    Mutex mtx;
    Mutex& outputMutex;
    Connection& conn;
    string_t host;
    string_t port;
};

/** Reconnection management using RAII. */
class ReconnectingGuard {
 public:
    ReconnectingGuard(ReconnectingConnection& rc);
    ~ReconnectingGuard();

 private:
    ReconnectingConnection& rc;

    ReconnectingGuard(const ReconnectingGuard&);
    ReconnectingGuard& operator=(const ReconnectingGuard&);
};

#endif
