/**
  *  \file planetscentral/userdir.h
  *  \brief File Server - User Directory Objects
  */
#ifndef C2SERVER_PLANETSCENTRAL_USERDIR_H
#define C2SERVER_PLANETSCENTRAL_USERDIR_H

#include <memory>
#include <map>
#include <time.h>
#include "cpluslib/ptrvector.h"
#include "cpluslib/string.h"
#include "cpluslib/types.h"
#include "io/streams.h"
#include "planetscentral/dirinfo.h"

class UserDirectory;

/** User directory object, base class. */
class UserItem {
 public:
    UserItem(string_t name, UserDirectory* parent);
    virtual ~UserItem();
    string_t getName() const;
    UserDirectory* getParent() const;
 private:
    string_t name;
    UserDirectory* parent;
};

/** User file. */
class UserFile : public UserItem {
    uint32_t size;
 public:
    UserFile(string_t name, UserDirectory* parent);
    ~UserFile();

    void setInfo(uint32_t size);
    uint32_t getSize() const;
};

/** User directory. */
class UserDirectory : public UserItem {
 public:
    enum {
        AllowRead   = 1,
        AllowWrite  = 2,
        AllowList   = 4,
        AllowAccess = 8,
        AllowAll    = AllowRead + AllowWrite + AllowList + AllowAccess
    };
    UserDirectory(string_t name, UserDirectory* parent);
    ~UserDirectory();

    bool wasRead() const;
    bool wasFullyRead() const;

    void read(string_t physical);
    void read();
    void readFully(string_t physical);
    void readFully();
    void forget();

    virtual string_t getFullPath();

    UserItem* find(string_t name);

    uint32_t getNumDirs() const;
    uint32_t getNumFiles() const;
    UserDirectory* getDir(uint32_t n);
    UserFile* getFile(uint32_t n);
    string_t getOwner() const;
    int getPermissions(string_t user) const;
    void setPermissions(string_t user, string_t perms);
    void listPermissions(std::vector<std::pair<string_t, string_t> >& perms);
    int getVisibilityLevel() const;

    string_t getProperty(string_t name) const;
    void setProperty(string_t name, string_t value);

    DirInfo& getDirInfo();

    bool createDirectory(string_t name, string_t forUser);
    bool createFile(string_t name, const string_t& data);
    bool removeEntry(UserItem* it);
    bool removeAllEntries();
    Ptr<Stream> openFile(string_t name, int mode);
    void computeTotals(int32_t& nfiles, int32_t& nkbytes);

 private:
    void loadControlFile(Directory& dir);
    void saveControlFile();
    void saveControlFile(Directory& dir);
    void setOwner();

    ptr_vector<UserDirectory> subdirs;
    ptr_vector<UserFile> files;
    bool was_read;
    uint32_t num_fully_read_subdirs;

    string_t owner;
    std::map<string_t, string_t> control_info;

    std::auto_ptr<DirInfo> dir_info;
};

/** Root directory. Differs from UserDirectory in that it does
    not have a parent. */
class UserRoot : public UserDirectory {
    string_t physical;
 public:
    UserRoot(string_t name, string_t physical);
    ~UserRoot();
    string_t getFullPath();
};

#endif
